from contextlib import contextmanager, nullcontext
from functools import partial
import logging
from importlib import import_module
import inspect
from mock import patch
from multiprocessing import Process, get_context
import os
from random import randrange
import requests
import signal
import sys
from threading import Thread
from typing import List, Optional
import time
import uuid


from asgi_correlation_id import correlation_id
import fastapi as fa
from prometheus_client import multiprocess, CollectorRegistry
from prometheus_fastapi_instrumentator import Instrumentator
from pydantic import BaseModel, Extra, Field
import uvicorn

from chaiverse.logging import get_service_logger_config
from chaiverse.fastapi_instrumentator import FastAPIInstrumentator

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


# Uvicorn uses spawn as the default multiprocessing method for
# compatibility with Windows. But this is slower and requires pickling.
# So we tell it to use fork instead.
uvicorn._subprocess.spawn = get_context("fork")


class ChaiverseServerGracefulShutdown(Exception):
    pass


class UvicornServer(uvicorn.Server):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = kwargs["config"]
        if self.config.workers == 1:
            signal.signal(signal.SIGTERM, self.graceful_shutdown)
            self.process = Thread(target=self.run)
        else:
            self.process = Process(target=self._multiprocess_target)

    def install_signal_handlers(self):
        if self.config.workers > 1:
            super().install_signal_handlers()

    @contextmanager
    def run_in_process(self):
        self.process.start()
        try:
            yield self
        finally:
            self.terminate()

    def terminate(self):
        self.should_exit = True
        if self.config.workers > 1:
            self.process.terminate()
        self.process.join()

    def graceful_shutdown(self, signum, frame):
        raise ChaiverseServerGracefulShutdown

    def _multiprocess_target(self):
        socket = self.config.bind_socket()
        multiprocess = uvicorn.supervisors.Multiprocess(
            config=self.config,
            target=self.run,
            sockets=[socket]
        )
        multiprocess.run()
        return multiprocess


class ChaiverseServerConfig(BaseModel):
    name: str = "chaiverse_server"
    port: int = 8080
    host: str = "0.0.0.0"
    workers: int = 1
    environment_variables: dict = Field(default_factory=dict)
    # Set to a positive number to limit the number of requests the uvicorn
    # process can handle. Used to avoid memory leaks. Random jitter is added
    # to ensure replicas don't shutdown at same time.
    limit_max_requests: int = -1
    log_level: str = "info"
    loki_base_url: str = ""
    prometheus_labels: Optional[dict] = {}
    # Specify the buckets to use when exporting route latencies to
    # prometheus histogram. This is a configurable option as we want
    # to avoid publishing too many time series to prometheus if we
    # don't need them
    prometheus_latency_buckets: List[float] = [0.1, 0.5, 1]
    prometheus_path: Optional[str] = None
    trace_header_name: Optional[str] = 'traceparent'

    class Config:
        extra = Extra.allow

    @classmethod
    def from_app(cls, app):
        config = cls._get_config_from_app(app)
        return cls(**config)

    @classmethod
    def from_request(cls, request: fa.Request):
        return cls.from_app(request.app)

    @classmethod
    def _get_config_from_app(cls, app: fa.FastAPI):
        config = getattr(app.state, "chaiverse_server_config", None)
        config = config.dict() if config else {}
        return config


class ChaiverseServer:
    def __init__(self, fastapi_app: fa.FastAPI, config):
        self.config = config
        self.fastapi_app_factory = partial(self._prepare_fastapi, fastapi_app)
        self.prometheus_instrumentator = None

    @contextmanager
    def serve(self):
        try:
            logger.info(f"Starting server {self.config.name}.")
            with self._start() as server:
                logger.info(f"Waiting for server {self.config.name} to be ready.")
                start = time.time()
                while not self._server_ready:
                    time.sleep(0.05)
                    pass
                duration = round(time.time() - start, 2)
                logger.info(f"Server {self.config.name} ready after {duration}s")
                yield server
        finally:
            logger.info(f"Shutting down server {self.config.name}.")

    @contextmanager
    def _start(self, **kwargs):
        config = {
            "app": self.fastapi_app_factory,
            "factory": True,
            "port": self.config.port,
            "host": self.config.host,
            "workers": self.config.workers,
            "log_config": self.logging_config,
            "log_level": self.config.log_level,
        }
        if self.config.limit_max_requests > 0:
            config.update({"limit_max_requests": self.limit_max_requests})
        config = uvicorn.Config(**config)
        server = UvicornServer(config=config)
        with self._environment_patch(), server.run_in_process():
            yield server

    @property
    def logging_config(self):
        return get_service_logger_config(self.config.name, self.config.loki_base_url)

    @property
    def _server_ready(self):
        try:
            requests.get(f"http://localhost:{self.config.port}/test_endpoint")
            is_ready = True
        except requests.exceptions.ConnectionError:
            logger.info(f"{self.config.name} not yet ready!")
            is_ready = False
        return is_ready

    def _prepare_fastapi(self, fastapi_app):
        # Install dependency solving patch (see comment in function)
        _patch_fastapi_dependency_solver()
        # Allow fastapi app to be passed in as import string
        # to enable specifying app via environment variable
        # in production
        if isinstance(fastapi_app, str):
            fastapi_app = _resolve_fastapi_app(fastapi_app)
        _install_general_exception_handler(fastapi_app)
        self.prometheus_instrumentator = FastAPIInstrumentator(fastapi_app, self.config)
        self.prometheus_instrumentator.instrument()

        fastapi_app.state.chaiverse_server_config = self.config
        return fastapi_app

    def _environment_patch(self):
        environment_patch = (
            patch.dict(os.environ, self.config.environment_variables)
            if self.config.environment_variables
            else nullcontext()
        )
        return environment_patch

    @property
    def limit_max_requests(self):
        base_limit = self.config.limit_max_requests
        adjusted_limit = randrange(base_limit, int(base_limit + 0.25 * base_limit))
        return adjusted_limit


def _install_general_exception_handler(fastapi_app):
    @fastapi_app.exception_handler(Exception)
    def runtime_exception_handler(request: fa.Request, exc: Exception):
        return exception_response(fastapi_app, 500, exc)


def exception_response(fastapi_app, status_code, exc):
    if (
        getattr(exc, "message", None)
        and "unhandled errors in a TaskGroup" in exc.message
    ):
        exc = exc.exceptions
    trace_header_name = fastapi_app.state.chaiverse_server_config.trace_header_name
    response = fa.responses.JSONResponse(
        status_code=status_code,
        content={"detail": str(exc)},
        headers={trace_header_name: correlation_id.get() or ""},
    )
    return response


def _resolve_fastapi_app(fastapi_app):
    import_path, app = fastapi_app.rsplit(".", 1)
    module = import_module(import_path)
    fastapi_app = getattr(module, app)
    return fastapi_app


def _patch_fastapi_dependency_solver():
    # FastAPI will resolve all dependencies that are not coroutines in a
    # separate thread (see solve_depdendencies in fastapi/depdendencies/utils.py)
    # However for low-overhead code such as class instantiation this is incredibly
    # wasteful at best, and at worst can lead to significant overhead in
    # latencies.
    # In particular, class-based views (which we call controllers in this
    # codebase) from fastapi_utils are treated as such dependencies, so
    # everytime a request hits a controller, there is an additional overhead
    # from a new thread being spun.
    # Here we patch fastapi to handle class depdenencies synchronously, and
    # avoid  this overhead.
    # See also https://github.com/fastapi/fastapi/discussions/7546 for
    # benchmarks showing the slowdown.
    from starlette.concurrency import run_in_threadpool as starlette_run_in_threadpool

    async def run_in_threadpool(call, **kwargs):
        if inspect.isclass(call):
            result = call(**kwargs)
        else:
            result = await starlette_run_in_threadpool(call, **kwargs)
        return result

    fa.dependencies.utils.run_in_threadpool = run_in_threadpool



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, wait, ALL_COMPLETED
from datetime import datetime
import hashlib
import inspect
import os
import pickle
from typing import Optional
from time import time

from typing_extensions import Literal

from tqdm import tqdm


CACHE_UPDATE_HOURS = 6


def get_guanaco_data_dir_env():
    home_dir = os.path.expanduser("~")
    return os.environ.get('GUANACO_DATA_DIR', f'{home_dir}/.chai-guanaco')


def guanaco_data_dir():
    data_dir = get_guanaco_data_dir_env()
    os.makedirs(os.path.join(data_dir, 'cache'), exist_ok=True)
    return data_dir


def print_color(text, color, **kwargs):
    colors = {
        'blue': '\033[94m',
        'cyan': '\033[96m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'red': '\033[91m',
        'white': '\033[97m',
        'yellow': '\033[93m',
        'magenta': '\033[95m',
        'grey': '\033[90m',
        'black': '\033[90m',
        'default': '\033[99m'
    }
    assert color in colors.keys()
    print(f'{colors[color]}{text}\033[0m', **kwargs)


def cache(func, regenerate=False):
    def wrapper(*args, **kwargs):
        file_path = _get_cache_file_path(func, args, kwargs)
        try:
            result = _load_from_cache(file_path)
            assert not regenerate
            # ensuring file is less than N hours old, otherwise regenerate
            assert (time() - os.path.getmtime(file_path)) < 3600 * CACHE_UPDATE_HOURS
        except (FileNotFoundError, AssertionError):
            result = func(*args, **kwargs)
            _save_to_cache(file_path, result)
        return result
    return wrapper


def _get_cache_file_path(func, args, kwargs):
    cache_dir = os.path.join(guanaco_data_dir(), 'cache')
    os.makedirs(cache_dir, exist_ok=True)
    func_name = func.__name__
    signature = _func_call_as_string(func, args, kwargs)
    signature_hexdigest = get_hexdigest(signature)
    fname = f'cache-{func_name}-{signature_hexdigest}'
    return os.path.join(cache_dir, f'{fname}.pkl')


def get_hexdigest(input_string):
    hexdigest = hashlib.md5(input_string.encode('UTF-8')).hexdigest()
    return hexdigest


def _load_from_cache(file_path):
    with open(file_path, 'rb') as f:
        return pickle.load(f)


def _save_to_cache(file_path, data):
    with open(file_path, 'wb') as f:
        pickle.dump(data, f)


def _func_call_as_string(func, args, kwargs):
    func_name = func.__name__
    param_names = list(inspect.signature(func).parameters.keys())
    arg_strs = [f'{name}={value!r}' for name, value in zip(param_names, args)]
    arg_strs += [f'{name}={value!r}' for name, value in kwargs.items()]
    return f"{func_name}({', '.join(arg_strs)})"


def get_localised_timestamp(timestamp, timezone=None):
    if not timezone:
        timezone = datetime.now().astimezone().tzinfo
    timestamp = datetime.fromisoformat(timestamp)
    timestamp = timestamp.astimezone(timezone)
    return timestamp


def parse_log_entry(log, timezone=None):
    timestamp = get_localised_timestamp(log["timestamp"], timezone)
    timestamp = timestamp.strftime("%H:%M:%S")
    message = [timestamp, log["level"], log["entry"]]
    message = ":".join(message)
    return message


def _distribute_to_multiple_workers(func, *args_iter, max_workers=2, worker_type: Literal['process', 'thread']='process', **kwargs):
    futures = []
    with tqdm(total=None) as progress:
        PoolExecutor = ProcessPoolExecutor if worker_type == 'process' else ThreadPoolExecutor
        with PoolExecutor(max_workers) as executor:
            for func_args in zip(*args_iter):
                future = executor.submit(func, *func_args, **kwargs)
                future.add_done_callback(lambda p: progress.update(1))
                futures.append(future)
            progress.total = len(futures)
            wait(futures, return_when=ALL_COMPLETED)
    results = [future.result() for future in futures]
    return results


def _distribute_to_single_worker(func, *args_iter, **kwargs):
    args_list = list(zip(*args_iter))
    results = [func(*args, **kwargs) for args in tqdm(args_list, total=len(args_list))]
    return results


def distribute_to_workers(func, *args_iter,  max_workers=1, worker_type: Literal['process', 'thread']='process', **kwargs):
    if max_workers == 1:
        return _distribute_to_single_worker(func, *args_iter, **kwargs)
    else:
        return _distribute_to_multiple_workers(func, *args_iter, max_workers=max_workers, worker_type=worker_type, **kwargs)



import contextlib
from mock import patch, Mock
import pkgutil
from typing import List, Optional

from fastapi import FastAPI
from fastapi.testclient import TestClient
from pydantic import BaseModel

from chaiverse.chaiverse_server import _resolve_fastapi_app, ChaiverseServerConfig
from chaiverse.database import MockDatabase
from chaiverse.fastapi_instrumentator import FastAPIInstrumentator


class BackendService(BaseModel):
    path: str
    url_patches: Optional[List] = []
    database_patches: Optional[List] = []
    code_patches: Optional[List] = []
    client_patches: Optional[List] = []

    @property
    def name(self):
        return self.path.split(".")[0]


# Mock background tasks as it can cause end to end tests
# to finish early (due to async behaviour)
class MockBackgroundTasks:
    def add_task(self, func, *args, **kwargs):
        return func(*args, **kwargs)


class MockThreadPoolExecutor():
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def submit(self, func, *args, **kwargs):
        return func(*args, **kwargs)


@contextlib.contextmanager
def run_servers(backend_services, mock_database=True, raise_server_exceptions=True, config=None):
    backend = FastAPI()
    config = config if config else ChaiverseServerConfig()
    # Mount all the microservices as subapps on the main fastapi app
    backend = mount_backends(backend, backend_services, config)
    # Patches to point to the launched backend
    client_patches = []
    mock_database = MockDatabase() if mock_database else None
    # Patches to install before launching the backend
    patches = get_backend_patches(backend_services, mock_database)
    with install_patches(patches), TestClient(backend, raise_server_exceptions=raise_server_exceptions) as client:
    # Patches to install that point to the test client
        client_patches = get_client_patches(backend_services, client)
        with install_patches(client_patches):
            yield client


def mount_backends(backend, backend_services, config):
    for service in backend_services:
        fastapi_app = _resolve_fastapi_app(service.path)
        fastapi_app.state.chaiverse_server_config = config
        instrumentator = FastAPIInstrumentator(fastapi_app, config)
        instrumentator.instrument()
        backend.mount(f"/{service.name}", fastapi_app)
    return backend


def get_client_patches(backend_services, client):
    client_patches = []
    for service in backend_services:
        service_client_patches = []
        for client_patch in service.client_patches:
            obj = pkgutil.resolve_name(client_patch)
            if callable(obj):
                client_patch = patch(client_patch, return_value=client)
            else:
                client_patch = patch(client_patch, client)
            client_patches.append(client_patch)
    return client_patches


def get_backend_patches(backend_services, mock_database):
    patches = []
    for service in backend_services:
        patches += get_service_patches(service, mock_database)
    return patches


def get_service_patches(service, mock_database):
    patches = []
    for url_patch in service.url_patches:
        url_patch = patch(url_patch, f"/{service.name}")
        patches.append(url_patch)
    if mock_database:
        for database_patch in service.database_patches:
            database_patch = patch(database_patch, mock_database)
            patches.append(database_patch)

    for code_patch in service.code_patches:
        patches.append(code_patch)
    return patches


@contextlib.contextmanager
def install_patches(patches):
    with contextlib.ExitStack() as stack:
        for mock_patch in patches:
            stack.enter_context(mock_patch)
        yield



import json
from typing import Optional

import firebase_admin
import firebase_admin.db
from flatten_dict import flatten

from chaiverse.database.inferno_database_adapter import _InfernoDatabaseAdapter


class _FirebaseDatabase(_InfernoDatabaseAdapter):
    def __init__(self, credentials_file: str, url: str):
        self.credentials_file = credentials_file
        self.url = url
        self._initialize_firebase()

    def reference(self, path: str):
        return firebase_admin.db.reference(path, app=self.app)

    def get(self, path: str, shallow: bool = False):
        firebase_reference = self.reference(path)
        record = firebase_reference.get(shallow=shallow)
        return record

    def is_in_database(self, path: str):
        value = self.get(path, shallow=True)
        return value is not None

    def set(self, path: str, value):
        firebase_reference = self.reference(path)
        value = serialise_input(value)
        firebase_reference.set(value)

    def update(self, path: str, record: dict):
        firebase_reference = self.reference(path)
        record = serialise_input(record)
        firebase_reference.update(record)

    def multi_update(self, path: str, record: dict):
        # While the logic for this function is the same as update, the intended
        # use-case is different due to the different syntax for multi-path
        # updates
        # See https://firebase.google.com/docs/database/admin/save-data#section-update
        record = flatten(record, reducer="path")
        self.update(path, record)

    def where(self, path, **kwargs):
        firebase_reference = self.reference(path)
        assert len(kwargs) > 0, "No query provided!"
        field_name, field_value = list(kwargs.items())[0]
        record = firebase_reference.order_by_child(field_name).equal_to(field_value).get()
        # Firebase doesn't support querying on multiple keys, so have to
        # filter the further keys here
        records = list(record.values())
        if len(kwargs) > 1:
            records = filter_records(records, kwargs)
        return records

    def remove(self, path):
        assert path not in ["", "/"]
        self.reference(path).delete()

    def query_by_key_range(self, path, start_at=None, end_at=None, limit_to_first: Optional[int] = None, limit_to_last: Optional[int] = None):
        firebase_reference = self.reference(path)
        query = firebase_reference.order_by_key()
        query = self._range_query_modifier(query, start_at, end_at, limit_to_first, limit_to_last)
        result = query.get()
        return result

    def query_by_child_value_range(self, path, by, start_at=None, end_at=None, limit_to_first: Optional[int] = None, limit_to_last: Optional[int] = None):
        firebase_reference = self.reference(path)
        query = firebase_reference.order_by_child(by)
        query = self._range_query_modifier(query, start_at, end_at, limit_to_first, limit_to_last)
        result = query.get()
        return result

    def atomic_add(self, path: str, value: float):
        func = partial(add, y=value)
        self.atomic_set(path, func)

    def atomic_increment(self, path: str):
        self.atomic_set(path, incrementer)

    def atomic_decrement(self, path: str):
        self.atomic_set(path, decrementer)

    def atomic_set(self, path: str, operation: callable):
        """Atomic version of set to avoid write clashes"""
        firebase_reference = self.reference(path)
        firebase_reference.transaction(operation)

    def _initialize_firebase(self):
        name = self.__class__.__name__
        try:
            credentials = firebase_admin.credentials.Certificate(self.credentials_file)
            init_params = {"databaseURL": self.url}
            self.app = firebase_admin.initialize_app(credentials, init_params, name=name)
        except ValueError:
            self.app = firebase_admin.get_app(name=name)

    def _range_query_modifier(self, query, start_at, end_at, limit_to_first, limit_to_last):
        if start_at:
            query = query.start_at(start_at)
        if end_at:
            query = query.end_at(end_at)
        if limit_to_first:
            query = query.limit_to_first(limit_to_first)
        if limit_to_last:
            query = query.limit_to_last(limit_to_last)
        return query


def serialise_input(record: dict):
    json_string = json.dumps(record, default=str)
    serialised_record = json.loads(json_string)
    return serialised_record


def add(x, y):
    return x + y


def incrementer(x):
    """Separate to make testable"""
    x = x if x else 0
    return x + 1


def decrementer(x):
    """Separate to make testable"""
    x = x if x else 0
    return x - 1


def filter_records(records, filters):
    filtered_records = []
    for record in records:
        # If filters is a subset of the record, then its good
        if filters.items() <= record.items():
            filtered_records.append(record)
    return filtered_records



from contextlib import contextmanager
import logging


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@contextmanager
def unsafe_swallow_exception(logging_label):
    try:
        yield logger
    except Exception as ex:
        logger.exception(f'CHAI_EXCEPTION:{logging_label} throws exception {repr(ex)}')
